from pyspark.sql.types import DataType


class Entity:
    def __init__(self, name: str, id_column: str, id_column_type: DataType, time_column: str = None, time_column_type: DataType = None):

        self.name = name
        self.id_column = id_column
        self.id_column_type = id_column_type
        self.time_column = time_column
        self.time_column_type = time_column_type



from daipecore.decorator.OutputDecorator import OutputDecorator
from injecta.container.ContainerInterface import ContainerInterface
from pyspark.sql import DataFrame
from featurestorebundle.entity.Entity import Entity
from featurestorebundle.feature.FeatureTemplateMatcher import FeatureTemplateMatcher
from featurestorebundle.feature.FeatureTemplate import FeatureTemplate
from featurestorebundle.feature.FeatureList import FeatureList
from featurestorebundle.feature.FeaturesStorage import FeaturesStorage
from featurestorebundle.metadata.MetadataHTMLDisplayer import MetadataHTMLDisplayer


class feature(OutputDecorator):  # noqa: N801
    def __init__(self, *args, entity: Entity, category: str = None, features_storage: FeaturesStorage = None):
        self._args = args
        self.__entity = entity
        self.__category = category
        self.__features_storage = features_storage

    def process_result(self, result: DataFrame, container: ContainerInterface):
        self.__check_primary_key_columns(result)

        if self.__features_storage:
            feature_template_matcher: FeatureTemplateMatcher = container.get(FeatureTemplateMatcher)

            feature_list = self.__prepare_features(feature_template_matcher, result, self._args)
            self.__features_storage.add(result, feature_list)

            metadata_html_displayer: MetadataHTMLDisplayer = container.get(MetadataHTMLDisplayer)
            metadata_html_displayer.display(feature_list.get_metadata_dicts())

    def __check_primary_key_columns(self, result: DataFrame):
        if self.__entity.id_column not in result.columns:
            raise Exception(f"{self.__entity.id_column} column is missing in the output dataframe")

        if self.__entity.time_column not in result.columns:
            raise Exception(f"{self.__entity.time_column} column is missing in the output dataframe")

    def __prepare_features(self, feature_template_matcher: FeatureTemplateMatcher, df: DataFrame, args: tuple) -> FeatureList:
        """
        @[foo]_feature(
            ("delayed_flights_pct_30d", "% of delayed flights in last 30 days"),
            ("early_flights_pct_30d", "% of flights landed ahead of time in last 30 days")
        )
        """
        feature_templates = [FeatureTemplate(*arg, category=self.__category) for arg in args]
        return feature_template_matcher.get_features(self.__entity, feature_templates, df)



from logging import Logger
from typing import List
from pyspark.sql import SparkSession

from featurestorebundle.feature.Feature import Feature
from featurestorebundle.feature.FeatureList import FeatureList


class FeatureManager:
    def __init__(self, logger: Logger, spark: SparkSession):
        self.__logger = logger
        self.__spark = spark

    def register(self, table_identifier: str, feature_list: FeatureList):
        def build_add_column_string(feature: Feature):
            return f'{feature.name} {feature.dtype} COMMENT "{feature.description}"'

        def build_add_columns_string(table_identifier, feature_list: FeatureList):
            add_column_sqls = [build_add_column_string(feature) for feature in feature_list.get_all()]
            return f"ALTER TABLE {table_identifier} ADD COLUMNS ({','.join(add_column_sqls)})"

        self.__spark.sql(build_add_columns_string(table_identifier, feature_list))

    def get_feature_names(self, table_identifier: str) -> List[str]:
        column_definitions = self.__spark.sql(f"DESCRIBE TABLE {table_identifier}").collect()

        def find_separation_row(column_definitions):
            for i, row in enumerate(column_definitions):
                if row.col_name == "" or row.col_name == "# Partition Information":
                    return i

            return None

        feature_rows = column_definitions[2 : find_separation_row(column_definitions)]  # noqa: E203
        return [row.col_name for row in feature_rows]

    def get_values(self, table_identifier: str, feature_names: List[str]):
        df = self.__spark.read.table(table_identifier)

        return df.select(df.columns[:2] + feature_names)



from pyspark.sql import SparkSession, DataFrame
from pyspark.sql import functions as f
from pyspark.sql import types as t
from delta.tables import DeltaTable
from featurestorebundle.entity.Entity import Entity
from featurestorebundle.feature.FeatureList import FeatureList
from gql import gql, Client
from logging import Logger

from featurestorebundle.metadata.MetadataWriter import MetadataWriter


class FeatureDataMerger:
    def __init__(
        self, metadata_api_enabled: bool, logger: Logger, gql_client: Client, spark: SparkSession, metadata_writer: MetadataWriter
    ):
        self.__metadata_api_enabled = metadata_api_enabled
        self.__logger = logger
        self.__gql_client = gql_client
        self.__spark = spark
        self.__metadata_writer = metadata_writer

    def merge(
        self,
        entity: Entity,
        feature_list: FeatureList,
        features_data: DataFrame,
        pk_columns: list,
        target_table_path: str,
        metadata_table_path: str,
        feature_store_type: str,
    ):
        feature_names = feature_list.get_names()

        data_column_names = [
            field.name for field in features_data.schema.fields if field.name not in [entity.id_column, entity.time_column]
        ]

        if len(data_column_names) != len(feature_names):
            raise Exception(
                f"Number or dataframe columns ({len(data_column_names)}) != number of features instances matched ({len(feature_names)})"
            )

        def build_update_set():
            update_set = {}

            for i, feature_instance_name in enumerate(feature_names):
                update_set[feature_instance_name] = f"source.{data_column_names[i]}"

            return update_set

        def build_insert_set():
            insert_set = build_update_set()
            insert_set[entity.id_column] = f"source.{entity.id_column}"
            insert_set[entity.time_column] = f"source.{entity.time_column}"

            return insert_set

        def build_merge_condition():
            conditions = []

            for pk in pk_columns:
                conditions.append(f"target.{pk} = source.{pk}")

            return " AND ".join(conditions)

        if feature_store_type == "delta":
            delta_table = DeltaTable.forPath(self.__spark, target_table_path)

            self.__logger.info(f"Writing feature data into {target_table_path}")

            (
                delta_table.alias("target")
                .merge(features_data.alias("source"), build_merge_condition())
                .whenMatchedUpdate(set=build_update_set())
                .whenNotMatchedInsert(values=build_insert_set())
                .execute()
            )

        if feature_store_type == "databricks":
            from databricks import feature_store  # noqa

            fs = feature_store.FeatureStoreClient()
            db_name = target_table_path.split(".")[0]

            self.__spark.sql(f"CREATE DATABASE IF NOT EXISTS {db_name}")

            final_df = features_data

            for field in features_data.schema:
                if isinstance(field.dataType, t.DecimalType):
                    if field.dataType.precision == 0:
                        final_df = final_df.withColumn(field.name, f.col(field.name).cast(t.LongType()))
                    else:
                        final_df = final_df.withColumn(field.name, f.col(field.name).cast(t.DoubleType()))

            try:
                fs.get_feature_table(target_table_path)
                feature_table_exist = True

            except Exception:
                feature_table_exist = False

            if not feature_table_exist:
                fs.create_feature_table(name=target_table_path, keys=pk_columns, schema=final_df.select(pk_columns).schema)

            fs.write_table(
                name=target_table_path,
                df=final_df,
                mode="merge",
            )

        self.__metadata_writer.write(metadata_table_path, feature_list)

    def __post_metadata_to_db(self, schema: t.StructType(), feature_list: FeatureList, entity: Entity):
        for field in schema[2:]:
            if field.name in feature_list.get_names():
                gql_query = gql(
                    f"""
                        mutation {{
                            createFeature(entity: "{entity.name}", name: "{field.name}", description: "{field.metadata.get('comment')}", category: "{field.metadata.get('category')}") {{
                                id,
                                existing,
                            }}
                        }}
                    """
                )

                try:
                    self.__gql_client.execute(gql_query)

                except BaseException:
                    self.__logger.warning("Cannot reach metadata api server. The metadata will not be written.")



class TableNames:
    def __init__(
        self,
        db_name_template: str,
        table_name_template: str,
        path_template: str,
        historized_path_template: str,
        archive_path_template: str,
    ):
        self.__db_name_template = db_name_template
        self.__table_name_template = table_name_template
        self.__latest_path_template = path_template
        self.__historized_path_template = historized_path_template
        self.__archive_path_template = archive_path_template

    def get_latest_table_name(self, entity_name):
        return self.__replace_placeholders(self.__table_name_template, entity_name) + "_latest"

    def get_historized_table_name(self, entity_name):
        return self.__replace_placeholders(self.__table_name_template, entity_name) + "_historized"

    def get_db_name(self, entity_name):
        return self.__replace_placeholders(self.__db_name_template, entity_name)

    def get_latest_table_identifier(self, entity_name):
        return f"delta.`{self.get_latest_path(entity_name)}`"

    def get_historized_table_identifier(self, entity_name):
        return f"delta.`{self.get_historized_path(entity_name)}`"

    def get_latest_path(self, entity_name: str) -> str:
        return self.__replace_placeholders(self.__latest_path_template, entity_name)

    def get_historized_path(self, entity_name: str) -> str:
        return self.__replace_placeholders(self.__historized_path_template, entity_name)

    def get_latest_metadata_path(self, entity_name: str) -> str:
        return self.__replace_placeholders(self.__latest_path_template.replace("{entity}", "{entity}_metadata"), entity_name)

    def get_historized_metadata_path(self, entity_name: str) -> str:
        return self.__replace_placeholders(self.__historized_path_template.replace("{entity}", "{entity}_metadata"), entity_name)

    def get_archive_path(self, entity_name: str, date: str) -> str:
        return self.__archive_path_template.format(entity=entity_name, date=date)

    def __get_tablename(self, entity_name: str) -> str:
        return self.__replace_placeholders(self.__table_name_template, entity_name)

    def __replace_placeholders(self, template: str, entity_name: str):
        return template.format(entity=entity_name)



from pyspark.sql import Column, functions as f
from typing import Any, List

from pyspark.sql import DataFrame

HOUR = 60 * 60
DAY = 24 * HOUR
WEEK = 7 * DAY

PERIODS = {
    "h": HOUR,
    "d": DAY,
    "w": WEEK,
}

time_window_column_template = "is_time_window_{time_window}"


def windowed(column: Column, time_window: str, default_value: Any = 0) -> Column:
    time_window_col_name = time_window_column_template.format(time_window=time_window)
    return f.when(f.col(time_window_col_name), column).otherwise(f.lit(default_value))


def with_time_windows(df: DataFrame, window_col: str, target_date_column: Column, time_windows: List) -> DataFrame:
    for time_window in time_windows:
        df = df.withColumn(
            time_window_column_template.format(time_window=time_window),
            f.col(window_col).cast("long") >= (target_date_column.cast("long") - PERIODS[time_window[-1]] * int(time_window[:-1])),
        )

    return df



from pyspark.sql import SparkSession
from pyspark.sql import DataFrame
from pyspark.sql import types as t
from featurestorebundle.entity.Entity import Entity


class EmptyDataFrameCreator:
    def __init__(self, spark: SparkSession):
        self.__spark = spark

    def create(self, entity: Entity) -> DataFrame:
        return self.__spark.createDataFrame(self.__spark.sparkContext.emptyRDD(), schema=self.__get_dataframe_schema(entity))

    def __get_dataframe_schema(self, entity: Entity) -> t.StructType():
        return t.StructType(
            [
                t.StructField(entity.id_column, entity.id_column_type, nullable=False),
                t.StructField(entity.time_column, entity.time_column_type, nullable=False),
            ]
        )



from typing import List
from pyspark.sql import SparkSession
from featurestorebundle.feature.FeatureManager import FeatureManager
from featurestorebundle.db.TableNames import TableNames


class FeatureStore:
    def __init__(
        self,
        spark: SparkSession,
        feature_manager: FeatureManager,
        table_names: TableNames,
    ):
        self.__spark = spark
        self.__feature_manager = feature_manager
        self.__table_names = table_names

    def get_latest(self, entity_name: str, feature_names: List[str] = None):
        table_identifier = self.__table_names.get_latest_table_identifier(entity_name)

        return self.__get_features(table_identifier, feature_names)

    def get_historized(self, entity_name: str, feature_names: List[str] = None):
        table_identifier = self.__table_names.get_historized_table_identifier(entity_name)

        return self.__get_features(table_identifier, feature_names)

    def __get_features(self, table_identifier: str, feature_names: List[str]):
        registered_feature_names_list = self.__feature_manager.get_feature_names(table_identifier)

        if feature_names is None:
            feature_names = registered_feature_names_list

        unregistered_features = set(feature_names) - set(registered_feature_names_list)

        if unregistered_features != set():
            raise Exception(f"Features {','.join(unregistered_features)} not registered")

        return self.__feature_manager.get_values(table_identifier, feature_names)



from logging import Logger
from pyspark.sql import DataFrame

from featurestorebundle.feature.FeatureStore import FeatureStore
from featurestorebundle.feature.FeatureDataMerger import FeatureDataMerger
from featurestorebundle.feature.FeaturesStorage import FeaturesStorage
from featurestorebundle.feature.TablePreparer import TablePreparer
from featurestorebundle.feature.FeaturesPreparer import FeaturesPreparer
from featurestorebundle.feature.writer.FeaturesWriterInterface import FeaturesWriterInterface
from featurestorebundle.db.TableNames import TableNames


class FeatureStoreWriter(FeaturesWriterInterface):
    def __init__(
        self,
        logger: Logger,
        feature_store: FeatureStore,
        table_preparer: TablePreparer,
        features_preparer: FeaturesPreparer,
        feature_data_merger: FeatureDataMerger,
        table_names: TableNames,
    ):
        self.__logger = logger
        self.__feature_store = feature_store
        self.__table_preparer = table_preparer
        self.__features_preparer = features_preparer
        self.__feature_data_merger = feature_data_merger
        self.__table_names = table_names

    def write_latest(self, features_storage: FeaturesStorage, archive=False):
        features_data = self.prepare_features(features_storage)
        feature_list = features_storage.feature_list
        entity = features_storage.entity

        table = f"{self.__table_names.get_db_name(entity.name)}.{self.__table_names.get_latest_table_name(entity.name)}"
        metadata_path = self.__table_names.get_latest_metadata_path(entity.name)
        pk_columns = [entity.id_column]
        feature_store_type = "databricks"

        self.__feature_data_merger.merge(
            entity,
            feature_list,
            features_data,
            pk_columns,
            table,
            metadata_path,
            feature_store_type,
        )

    def write_historized(self, features_storage: FeaturesStorage):
        features_data = self.prepare_features(features_storage)
        feature_list = features_storage.feature_list
        entity = features_storage.entity

        table = f"{self.__table_names.get_db_name(entity.name)}.{self.__table_names.get_historized_table_name(entity.name)}"
        metadata_path = self.__table_names.get_historized_metadata_path(entity.name)
        pk_columns = [entity.id_column, entity.time_column]
        feature_store_type = "databricks"

        self.__feature_data_merger.merge(
            entity,
            feature_list,
            features_data,
            pk_columns,
            table,
            metadata_path,
            feature_store_type,
        )

    def prepare_features(self, features_storage: FeaturesStorage) -> DataFrame:
        return self.__features_preparer.prepare(features_storage)



from logging import Logger

from delta.tables import DeltaTable
from pyspark.sql import SparkSession
from pyspark.sql import types as t

from featurestorebundle.feature.FeatureList import FeatureList

METADATA_PK = t.StructField("feature", t.StringType(), True)
METADATA_COLUMNS = [
    t.StructField("description", t.StringType(), True),
    t.StructField("extra", t.MapType(t.StringType(), t.StringType(), True)),
    t.StructField("feature_template", t.StringType(), True),
    t.StructField("category", t.StringType(), True),
    t.StructField("dtype", t.StringType(), True),
]
METADATA_SCHEMA = t.StructType([METADATA_PK] + METADATA_COLUMNS)


class MetadataWriter:
    def __init__(self, logger: Logger, spark: SparkSession):
        self.__logger = logger
        self.__spark = spark

    def write(self, metadata_path: str, feature_list: FeatureList):
        metadata_df = self.__spark.createDataFrame(feature_list.get_metadata(), METADATA_SCHEMA)

        if not DeltaTable.isDeltaTable(self.__spark, metadata_path):
            metadata_df.write.format("delta").save(metadata_path)
            return

        delta_table = DeltaTable.forPath(self.__spark, metadata_path)

        update_set = {col.name: f"source.{col.name}" for col in METADATA_COLUMNS}

        insert_set = {METADATA_PK.name: f"source.{METADATA_PK.name}", **update_set}

        self.__logger.info(f"Writing metadata into {metadata_path}")

        (
            delta_table.alias("target")
            .merge(metadata_df.alias("source"), "target.feature = source.feature")
            .whenMatchedUpdate(set=update_set)
            .whenNotMatchedInsert(values=insert_set)
            .execute()
        )



from pyspark.sql import SparkSession
from featurestorebundle.entity.Entity import Entity


class EmptyTableCreator:
    def __init__(self, spark: SparkSession):
        self.__spark = spark

    def create(self, table_name: str, path: str, entity: Entity):
        def build_create_entity_table_string(entity: Entity):
            return (
                f"CREATE TABLE IF NOT EXISTS {table_name}\n"
                f'({entity.id_column} {entity.id_column_type.typeName()} NOT NULL COMMENT "Entity id column",\n'
                f'{entity.time_column} {entity.time_column_type.typeName()} NOT NULL COMMENT "Compute time column")\n'
                f"USING DELTA\n"
                f"LOCATION '{path}'\n"
                f'COMMENT "The table contains entity {entity.name} features"\n'
            )

        self.__spark.sql(build_create_entity_table_string(entity))



from typing import List
from injecta.container.ContainerInterface import ContainerInterface
from daipecore.function.input_decorator_function import input_decorator_function
from featurestorebundle.feature.FeatureStore import FeatureStore


@input_decorator_function
def read_features(entity_name: str, feature_names: List[str] = None, historized: bool = False):
    def wrapper(container: ContainerInterface):
        feature_store: FeatureStore = container.get(FeatureStore)

        if historized:
            return feature_store.get_historized(entity_name, feature_names)
        return feature_store.get_latest(entity_name, feature_names)

    return wrapper



from pyspark.sql import SparkSession
from featurestorebundle.db.TableNames import TableNames
from featurestorebundle.entity.Entity import Entity
from featurestorebundle.feature.EmptyFileCreator import EmptyFileCreator
from featurestorebundle.feature.FeatureList import FeatureList
from featurestorebundle.feature.FeatureManager import FeatureManager


class TablePreparer:
    def __init__(
        self,
        table_names: TableNames,
        spark: SparkSession,
        empty_file_creator: EmptyFileCreator,
        features_manager: FeatureManager,
    ):
        self.__table_names = table_names
        self.__spark = spark
        self.__empty_file_creator = empty_file_creator
        self.__features_manager = features_manager

    def prepare(self, table_identifier: str, path: str, entity: Entity, current_feature_list: FeatureList):
        self.__empty_file_creator.create(path, entity)
        self.__register_features(table_identifier, current_feature_list)

    def __register_features(self, table_identifier: str, current_feature_list: FeatureList):
        registered_feature_names = self.__features_manager.get_feature_names(table_identifier)
        unregistered_features = current_feature_list.get_unregistered(registered_feature_names)

        if not unregistered_features.empty():
            self.__features_manager.register(table_identifier, unregistered_features)



from pyspark.sql import SparkSession
from delta.tables import DeltaTable
from featurestorebundle.entity.Entity import Entity
from featurestorebundle.feature.EmptyDataFrameCreator import EmptyDataFrameCreator


class EmptyFileCreator:
    def __init__(self, spark: SparkSession, empty_dataframe_creator: EmptyDataFrameCreator):
        self.__spark = spark
        self.__empty_dataframe_creator = empty_dataframe_creator

    def create(self, path: str, entity: Entity):
        if not DeltaTable.isDeltaTable(self.__spark, path):
            df = self.__empty_dataframe_creator.create(entity)
            df.write.format("delta").save(path)



PERIODS = {
    "h": "hours",
    "d": "days",
    "w": "weeks",
}


class TimeWindowHandler:
    def is_valid(self, time_window: str, feature_name: str):
        if not time_window[:-1].isdigit():
            raise Exception(f"In feature '{feature_name}', time_window={time_window[:-1]} is not a positive integer.")
        elif not time_window[-1] in PERIODS.keys():
            raise Exception(f"In feature '{feature_name}', time_window period '{time_window[-1]}' is not from {', '.join(PERIODS.keys())}")

    def to_text(self, time_window: str) -> str:
        n = int(time_window[:-1])
        period = time_window[-1]

        result = f"{n} {PERIODS[period]}"
        return result[:-1] if n == 1 else result



import datetime as dt
from logging import Logger
from pyspark.sql import DataFrame

from featurestorebundle.feature.FeatureStore import FeatureStore
from featurestorebundle.feature.FeatureDataMerger import FeatureDataMerger
from featurestorebundle.feature.FeaturesStorage import FeaturesStorage
from featurestorebundle.feature.TablePreparer import TablePreparer
from featurestorebundle.feature.FeaturesPreparer import FeaturesPreparer
from featurestorebundle.feature.writer.FeaturesWriterInterface import FeaturesWriterInterface
from featurestorebundle.db.TableNames import TableNames


class DeltaWriter(FeaturesWriterInterface):
    def __init__(
        self,
        logger: Logger,
        feature_store: FeatureStore,
        table_preparer: TablePreparer,
        features_preparer: FeaturesPreparer,
        feature_data_merger: FeatureDataMerger,
        table_names: TableNames,
    ):
        self.__logger = logger
        self.__feature_store = feature_store
        self.__table_preparer = table_preparer
        self.__features_preparer = features_preparer
        self.__feature_data_merger = feature_data_merger
        self.__table_names = table_names

    def write_latest(self, features_storage: FeaturesStorage, archive=False):
        features_data = self.prepare_features(features_storage)
        feature_list = features_storage.feature_list
        entity = features_storage.entity

        table_identifier = self.__table_names.get_latest_table_identifier(entity.name)
        path = self.__table_names.get_latest_path(entity.name)
        metadata_path = self.__table_names.get_latest_metadata_path(entity.name)
        pk_columns = [entity.id_column]
        feature_store_type = "delta"

        if archive:
            today_str = dt.date.today().strftime("%Y-%m-%d")
            archive_path = self.__table_names.get_archive_path(entity.name, today_str)
            feature_store_df = self.__feature_store.get_latest(entity.name)
            feature_store_df.write.format("delta").save(archive_path)

        self.__table_preparer.prepare(table_identifier, path, entity, feature_list)

        self.__feature_data_merger.merge(
            entity,
            feature_list,
            features_data,
            pk_columns,
            path,
            metadata_path,
            feature_store_type,
        )

    def write_historized(self, features_storage: FeaturesStorage):
        features_data = self.prepare_features(features_storage)
        feature_list = features_storage.feature_list
        entity = features_storage.entity

        table_identifier = self.__table_names.get_historized_table_identifier(entity.name)
        path = self.__table_names.get_historized_path(entity.name)
        metadata_path = self.__table_names.get_historized_metadata_path(entity.name)
        pk_columns = [entity.id_column, entity.time_column]
        feature_store_type = "delta"

        self.__table_preparer.prepare(table_identifier, path, entity, feature_list)

        self.__feature_data_merger.merge(
            entity,
            feature_list,
            features_data,
            pk_columns,
            path,
            metadata_path,
            feature_store_type,
        )

    def prepare_features(self, features_storage: FeaturesStorage) -> DataFrame:
        return self.__features_preparer.prepare(features_storage)
